@@ -6,7 +6,7 @@
  * Copyright (C) 1998 Raph Levien, 
  * 	     (C) 2001 ALTLinux, Moscow (http://www.alt-linux.org), 
  *           (C) 2001 Peter Novodvorsky (nidd@cs.msu.su)
- *           (C) 2006 László Németh (nemeth at OOo)
+ *           (C) 2006, 2007, 2008, 2010 László Németh (nemeth at OOo)
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -46,6 +46,18 @@
 
 #define noVERBOSE
 
+/* calculate hyphenmin values with long ligature length (2 or 3 characters
+ * instead of 1 or 2) for comparison with hyphenation without ligatures */
+#define noLONG_LIGATURE
+
+#ifdef LONG_LIGATURE
+#define LIG_xx	1
+#define LIG_xxx	2
+#else
+#define LIG_xx	0
+#define LIG_xxx	1
+#endif
+
 #include "hnjalloc.h"
 #include "hyphen.h"
 
@@ -233,7 +245,7 @@ get_state_str (int state)
 HyphenDict *
 hnj_hyphen_load (const char *fn)
 {
-  HyphenDict *dict;
+  HyphenDict *dict[2];
   HashTab *hashtab;
   FILE *f;
   char buf[MAX_CHARS];
@@ -243,42 +255,70 @@ hnj_hyphen_load (const char *fn)
   signed char replindex;
   signed char replcut;
   int state_num = 0, last_state;
-  int i, j;
+  int i, j, k;
   char ch;
   int found;
   HashEntry *e;
+  int nextlevel = 0;
 
   f = fopen (fn, "r");
   if (f == NULL)
     return NULL;
 
+// loading one or two dictionaries (separated by NEXTLEVEL keyword)
+for (k = 0; k == 0 || (k == 1 && nextlevel); k++) { 
   hashtab = hnj_hash_new ();
 #ifdef VERBOSE
   global = hashtab;
 #endif
   hnj_hash_insert (hashtab, "", 0);
-
-  dict = hnj_malloc (sizeof(HyphenDict));
-  dict->num_states = 1;
-  dict->states = hnj_malloc (sizeof(HyphenState));
-  dict->states[0].match = NULL;
-  dict->states[0].repl = NULL;
-  dict->states[0].fallback_state = -1;
-  dict->states[0].num_trans = 0;
-  dict->states[0].trans = NULL;
+  dict[k] = hnj_malloc (sizeof(HyphenDict));
+  dict[k]->num_states = 1;
+  dict[k]->states = hnj_malloc (sizeof(HyphenState));
+  dict[k]->states[0].match = NULL;
+  dict[k]->states[0].repl = NULL;
+  dict[k]->states[0].fallback_state = -1;
+  dict[k]->states[0].num_trans = 0;
+  dict[k]->states[0].trans = NULL;
+  dict[k]->nextlevel = NULL;
+  dict[k]->lhmin = 0;
+  dict[k]->rhmin = 0;
+  dict[k]->clhmin = 0;
+  dict[k]->crhmin = 0;
 
   /* read in character set info */
-  for (i=0;i<MAX_NAME;i++) dict->cset[i]= 0;
-  fgets(dict->cset,  sizeof(dict->cset),f);
-  for (i=0;i<MAX_NAME;i++)
-    if ((dict->cset[i] == '\r') || (dict->cset[i] == '\n'))
-        dict->cset[i] = 0;
-  dict->utf8 = (strcmp(dict->cset, "UTF-8") == 0);
+  if (k == 0) {
+    for (i=0;i<MAX_NAME;i++) dict[k]->cset[i]= 0;
+    fgets(dict[k]->cset,  sizeof(dict[k]->cset),f);
+    for (i=0;i<MAX_NAME;i++)
+      if ((dict[k]->cset[i] == '\r') || (dict[k]->cset[i] == '\n'))
+        dict[k]->cset[i] = 0;
+    dict[k]->utf8 = (strcmp(dict[k]->cset, "UTF-8") == 0);
+  } else {
+    strcpy(dict[k]->cset, dict[0]->cset);
+    dict[k]->utf8 = dict[0]->utf8;
+  }
 
   while (fgets (buf, sizeof(buf), f) != NULL)
     {
       if (buf[0] != '%')
 	{
+	  if (strncmp(buf, "NEXTLEVEL", 9) == 0) {
+	    nextlevel = 1;
+	    break;
+	  } else if (strncmp(buf, "LEFTHYPHENMIN", 13) == 0) {
+	    dict[k]->lhmin = atoi(buf + 13);
+	    continue;
+	  } else if (strncmp(buf, "RIGHTHYPHENMIN", 14) == 0) {
+	    dict[k]->rhmin = atoi(buf + 14);
+	    continue;
+	  } else if (strncmp(buf, "COMPOUNDLEFTHYPHENMIN", 21) == 0) {
+	    dict[k]->clhmin = atoi(buf + 21);
+	    continue;
+	  } else if (strncmp(buf, "COMPOUNDRIGHTHYPHENMIN", 22) == 0) {
+	    dict[k]->crhmin = atoi(buf + 22);
+	    continue;
+	  } 
 	  j = 0;
 	  pattern[j] = '0';
           repl = strchr(buf, '/');
@@ -298,7 +338,7 @@ hnj_hyphen_load (const char *fn)
             } else {
                 hnj_strchomp(repl + 1);
                 replindex = 0;
-                replcut = strlen(buf);
+                replcut = (signed char) strlen(buf);
             }
             repl = hnj_strdup(repl + 1);
           }
@@ -322,7 +362,7 @@ hnj_hyphen_load (const char *fn)
           } else {
             if (*word == '.') i++;
             /* convert UTF-8 char. positions of discretionary hyph. replacements to 8-bit */
-            if (dict->utf8) {
+            if (dict[k]->utf8) {
                 int pu = -1;        /* unicode character position */
                 int ps = -1;        /* unicode start position (original replindex) */
                 int pc = (*word == '.') ? 1: 0; /* 8-bit character position */
@@ -331,10 +371,10 @@ hnj_hyphen_load (const char *fn)
                     if ((((unsigned char) word[pc]) >> 6) != 2) pu++;
                     if ((ps < 0) && (replindex == pu)) {
                         ps = replindex;
-                        replindex = pc;
+                        replindex = (signed char) pc;
                     }
                     if ((ps >= 0) && ((pu - ps) == replcut)) {
-                        replcut = (pc - replindex);
+                        replcut = (signed char) (pc - replindex);
                         break;
                     }
                 }
@@ -346,14 +386,14 @@ hnj_hyphen_load (const char *fn)
 	  printf ("word %s pattern %s, j = %d  repl: %s\n", word, pattern + i, j, repl);
 #endif
 	  found = hnj_hash_lookup (hashtab, word);
-	  state_num = hnj_get_state (dict, hashtab, word);
-	  dict->states[state_num].match = hnj_strdup (pattern + i);
-	  dict->states[state_num].repl = repl;
-	  dict->states[state_num].replindex = replindex;
+	  state_num = hnj_get_state (dict[k], hashtab, word);
+	  dict[k]->states[state_num].match = hnj_strdup (pattern + i);
+	  dict[k]->states[state_num].repl = repl;
+	  dict[k]->states[state_num].replindex = replindex;
           if (!replcut) {
-            dict->states[state_num].replcut = strlen(word);
+            dict[k]->states[state_num].replcut = (signed char) strlen(word);
           } else {
-            dict->states[state_num].replcut = replcut;
+            dict[k]->states[state_num].replcut = replcut;
           }
 
 	  /* now, put in the prefix transitions */
@@ -363,8 +403,8 @@ hnj_hyphen_load (const char *fn)
 	      ch = word[j - 1];
 	      word[j - 1] = '\0';
 	      found = hnj_hash_lookup (hashtab, word);
-	      state_num = hnj_get_state (dict, hashtab, word);
-	      hnj_add_trans (dict, state_num, last_state, ch);
+	      state_num = hnj_get_state (dict[k], hashtab, word);
+	      hnj_add_trans (dict[k], state_num, last_state, ch);
 	    }
 	}
     }
@@ -399,25 +439,28 @@ hnj_hyphen_load (const char *fn)
 	  }
         /* KBH: FIXME state 0 fallback_state should always be -1? */
 	if (e->val)
-	  dict->states[e->val].fallback_state = state_num;
+	  dict[k]->states[e->val].fallback_state = state_num;
       }
 #ifdef VERBOSE
   for (i = 0; i < HASH_SIZE; i++)
     for (e = hashtab->entries[i]; e; e = e->next)
       {
 	printf ("%d string %s state %d, fallback=%d\n", i, e->key, e->val,
-		dict->states[e->val].fallback_state);
-	for (j = 0; j < dict->states[e->val].num_trans; j++)
-	  printf (" %c->%d\n", dict->states[e->val].trans[j].ch,
-		  dict->states[e->val].trans[j].new_state);
+		dict[k]->states[e->val].fallback_state);
+	for (j = 0; j < dict[k]->states[e->val].num_trans; j++)
+	  printf (" %c->%d\n", dict[k]->states[e->val].trans[j].ch,
+		  dict[k]->states[e->val].trans[j].new_state);
       }
 #endif
 
 #ifndef VERBOSE
   hnj_hash_free (hashtab);
 #endif
-
-  return dict;
+  state_num = 0;
+}
+  fclose(f);
+  if (k == 2) dict[0]->nextlevel = dict[1];
+  return dict[0];
 }
 
 void hnj_hyphen_free (HyphenDict *dict)
@@ -435,6 +478,7 @@ void hnj_hyphen_free (HyphenDict *dict)
       if (hstate->trans)
 	hnj_free (hstate->trans);
     }
+  if (dict->nextlevel) hnj_hyphen_free(dict->nextlevel);
 
   hnj_free (dict->states);
 
@@ -463,16 +507,16 @@ int hnj_hyphen_hyphenate (HyphenDict *dict,
 
   j = 0;
   prep_word[j++] = '.';
-  
+
   for (i = 0; i < word_size; i++)
       prep_word[j++] = word[i];
 
   prep_word[j++] = '.';
   prep_word[j] = '\0';
-      
-  for (i = 0; i < j; i++)                                                       
-    hyphens[i] = '0';    
-  
+
+  for (i = 0; i < word_size + 5; i++)
+    hyphens[i] = '0';
+
 #ifdef VERBOSE
   printf ("prep_word = %s\n", prep_word);
 #endif
@@ -563,12 +607,100 @@ int hnj_hyphen_hyphenate (HyphenDict *dict,
 
   if (prep_word != prep_word_buf)
     hnj_free (prep_word);
+    
   return 0;    
 }
 
-int hnj_hyphen_hyphenate2 (HyphenDict *dict,
-			   const char *word, int word_size, char * hyphens,
-			   char *hyphword, char *** rep, int ** pos, int ** cut)
+/* Unicode ligature length */
+int hnj_ligature(unsigned char c) {
+    switch (c) {
+        case 0x80:			/* ff */
+        case 0x81:			/* fi */
+        case 0x82: return LIG_xx;	/* fl */
+        case 0x83:			/* ffi */
+        case 0x84: return LIG_xxx;	/* ffl */
+        case 0x85:			/* long st */
+        case 0x86: return LIG_xx;	/* st */
+    }
+    return 0;
+}
+
+/* character length of the first n byte of the input word */
+int hnj_hyphen_strnlen(const char * word, int n, int utf8)
+{
+    int i = 0;
+    int j = 0;
+    while (j < n && word[j] != '\0') {
+      i++;
+      // Unicode ligature support
+      if (utf8 && ((unsigned char) word[j] == 0xEF) && ((unsigned char) word[j + 1] == 0xAC))  {
+        i += hnj_ligature(word[j + 2]);
+      }
+      for (j++; utf8 && (word[j] & 0xc0) == 0x80; j++);
+    }
+    return i;
+}
+
+int hnj_hyphen_lhmin(int utf8, const char *word, int word_size, char * hyphens,
+	char *** rep, int ** pos, int ** cut, int lhmin)
+{
+    int i = 1, j;
+
+    // Unicode ligature support
+    if (utf8 && ((unsigned char) word[0] == 0xEF) && ((unsigned char) word[1] == 0xAC))  {
+      i += hnj_ligature(word[2]);
+    }
+
+    for (j = 0; i < lhmin && word[j] != '\0'; i++) do {
+      // check length of the non-standard part
+      if (*rep && *pos && *cut && (*rep)[j]) {
+        char * rh = strchr((*rep)[j], '=');
+        if (rh && (hnj_hyphen_strnlen(word, j - (*pos)[j] + 1, utf8) +
+          hnj_hyphen_strnlen((*rep)[j], rh - (*rep)[j], utf8)) < lhmin) {
+            free((*rep)[j]);
+            (*rep)[j] = NULL;
+            hyphens[j] = '0';
+          }
+       } else {
+         hyphens[j] = '0';
+       }
+       j++;
+
+       // Unicode ligature support
+       if (utf8 && ((unsigned char) word[j] == 0xEF) && ((unsigned char) word[j + 1] == 0xAC))  {
+         i += hnj_ligature(word[j + 2]);
+       }
+    } while (utf8 && (word[j] & 0xc0) == 0x80);
+    return 0;
+}
+
+int hnj_hyphen_rhmin(int utf8, const char *word, int word_size, char * hyphens,
+	char *** rep, int ** pos, int ** cut, int rhmin)
+{
+    int i;
+    int j = word_size - 2;    
+    for (i = 1; i < rhmin && j > 0; j--) {
+      // check length of the non-standard part
+      if (*rep && *pos && *cut && (*rep)[j]) {
+        char * rh = strchr((*rep)[j], '=');
+        if (rh && (hnj_hyphen_strnlen(word + j - (*pos)[j] + (*cut)[j] + 1, 100, utf8) +
+          hnj_hyphen_strnlen(rh + 1, strlen(rh + 1), utf8)) < rhmin) {
+            free((*rep)[j]);
+            (*rep)[j] = NULL;
+            hyphens[j] = '0';
+          }
+       } else {
+         hyphens[j] = '0';
+       }
+       if (!utf8 || (word[j] & 0xc0) != 0xc0) i++;
+    }
+    return 0;
+}
+
+// recursive function for compound level hyphenation
+int hnj_hyphen_hyph_(HyphenDict *dict, const char *word, int word_size,
+    char * hyphens, char *** rep, int ** pos, int ** cut,
+    int clhmin, int crhmin, int lend, int rend)
 {
   char prep_word_buf[MAX_WORD];
   char *prep_word;
@@ -709,26 +841,19 @@ int hnj_hyphen_hyphenate2 (HyphenDict *dict,
   putchar ('\n');
 #endif
 
-  for (i = 0; i < j - 4; i++)
+  for (i = 0; i < j - 3; i++)
 #if 0
     if (hyphens[i + 1] & 1)
       hyphens[i] = '-';
 #else
     hyphens[i] = hyphens[i + 1];
 #endif
-  hyphens[0] = '0';
   for (; i < word_size; i++)
     hyphens[i] = '0';
   hyphens[word_size] = '\0';
 
-
-  if (prep_word != prep_word_buf) {
-    hnj_free (prep_word);
-  }
-
-        
        /* now create a new char string showing hyphenation positions */
-       /* count the hyphens and allocate space for the new hypehanted string */
+       /* count the hyphens and allocate space for the new hyphenated string */
        nHyphCount = 0;
        for (i = 0; i < word_size; i++)
           if (hyphens[i]&1)
@@ -752,34 +877,119 @@ int hnj_hyphen_hyphenate2 (HyphenDict *dict,
                     (*pos)[matchindex[i] - 1] = matchindex[i] - i;
                     (*cut)[matchindex[i] - 1] = matchlen[i];
                 }
-                if (hyphword) strcpy(hyphword + j, matchrepl[matchindex[i]]);
                 j += strlen(matchrepl[matchindex[i]]);
                 i += matchlen[i] - 1;
-          } else if (hyphword) {
-            hyphword[j++] = word[i];
-            if ((hyphens[i]&1) && !(isrepl && ((i+1) < word_size) && 
-                (matchindex[i+1] >= 0) && matchrepl[matchindex[i+1]])) hyphword[j++] = '=';
           }
        }
 
-  if (hyphword) hyphword[j] = '\0';
-
   if (matchrepl != matchrepl_buf) {
     hnj_free (matchrepl);
     hnj_free (matchlen);
     hnj_free (matchindex);
   }
 
-  if (!(dict->utf8)) return 0;
+  // recursive hyphenation of the first (compound) level segments
+  if (dict->nextlevel) {
+     char * rep2_buf[MAX_WORD];
+     int pos2_buf[MAX_WORD];
+     int cut2_buf[MAX_WORD];
+     char hyphens2_buf[MAX_WORD];
+     char ** rep2;
+     int * pos2;
+     int * cut2;
+     char * hyphens2;
+     int begin = 0;
+     if (word_size < MAX_CHARS) {
+        rep2 = rep2_buf;
+        pos2 = pos2_buf;
+        cut2 = cut2_buf;
+        hyphens2 = hyphens2_buf;
+     } else {
+        rep2 = hnj_malloc (word_size * sizeof(char *));
+        pos2 = hnj_malloc (word_size * sizeof(int));
+        cut2 = hnj_malloc (word_size * sizeof(int));
+        hyphens2 = hnj_malloc (word_size);
+     }
+     for (i = 0; i < word_size; i++) rep2[i] = NULL;
+     for (i = 0; i < word_size; i++) if 
+        (hyphens[i]&1 || (begin > 0 && i + 1 == word_size)) {
+        if (i - begin > 1) {
+            int hyph = 0;
+            prep_word[i + 2] = '\0';
+            /* non-standard hyphenation at compound boundary (Schiffahrt) */
+            if (*rep && *pos && *cut && (*rep)[i]) {
+                char * l = strchr((*rep)[i], '=');
+                strcpy(prep_word + 2 + i - (*pos)[i], (*rep)[i]);
+                if (l) {
+                    hyph = (l - (*rep)[i]) - (*pos)[i];
+                    prep_word[2 + i + hyph] = '\0';
+                }
+            }
+            hnj_hyphen_hyph_(dict, prep_word + begin + 1, i - begin + 1 + hyph,
+                hyphens2, &rep2, &pos2, &cut2, clhmin,
+                crhmin, (begin > 0 ? 0 : lend), (hyphens[i]&1 ? 0 : rend));
+            for (j = 0; j < i - begin - 1; j++) {
+                hyphens[begin + j] = hyphens2[j];
+                if (rep2[j] && rep && pos && cut) {
+                    if (!*rep && !*pos && !*cut) {
+                        int k;
+                        *rep = (char **) malloc(sizeof(char *) * word_size);
+                        *pos = (int *) malloc(sizeof(int) * word_size);
+                        *cut = (int *) malloc(sizeof(int) * word_size);
+                        for (k = 0; k < word_size; k++) {
+                            (*rep)[k] = NULL;
+                            (*pos)[k] = 0;
+                            (*cut)[k] = 0;
+                        }
+                    }
+                    (*rep)[begin + j] = rep2[j];
+                    (*pos)[begin + j] = pos2[j];
+                    (*cut)[begin + j] = cut2[j];
+                }
+            }
+            prep_word[i + 2] = word[i + 1];
+            if (*rep && *pos && *cut && (*rep)[i]) {
+                strcpy(prep_word + 1, word);
+            }
+        }
+        begin = i + 1;
+        for (j = 0; j < word_size; j++) rep2[j] = NULL;
+     }
+     
+     // non-compound
+     if (begin == 0) {
+        hnj_hyphen_hyph_(dict->nextlevel, word, word_size,
+            hyphens, rep, pos, cut, clhmin, crhmin, lend, rend);
+        if (!lend) hnj_hyphen_lhmin(dict->utf8, word, word_size, hyphens,
+            rep, pos, cut, clhmin);
+        if (!rend) hnj_hyphen_rhmin(dict->utf8, word, word_size, hyphens,
+            rep, pos, cut, crhmin);
+     }
+     
+     if (rep2 != rep2_buf) {
+        free(rep2);
+        free(cut2);
+        free(pos2);
+        free(hyphens2);
+     }
+  }
 
+  if (prep_word != prep_word_buf) hnj_free (prep_word);
+  return 0;
+}
+
+/* UTF-8 normalization of hyphen and non-standard positions */
+int hnj_hyphen_norm(const char *word, int word_size, char * hyphens,
+	char *** rep, int ** pos, int ** cut)
+{
+  int i, j, k;
   if ((((unsigned char) word[0]) >> 6) == 2) {
     fprintf(stderr, "error - bad, non UTF-8 input: %s\n", word);
     return 1;
   }
 
   /* calculate UTF-8 character positions */
-  j = -1;
-  for (i = 0; i < word_size; i++) {
+  for (i = 0, j = -1; i < word_size; i++) {
     /* beginning of an UTF-8 character (not '10' start bits) */
     if ((((unsigned char) word[i]) >> 6) != 2) j++;
     hyphens[j] = hyphens[i];
@@ -806,3 +1016,57 @@ int hnj_hyphen_hyphenate2 (HyphenDict *dict,
   hyphens[j + 1] = '\0';
   return 0;
 }
+
+/* get the word with all possible hyphenations (output: hyphword) */
+void hnj_hyphen_hyphword(const char * word, int l, const char * hyphens, 
+    char * hyphword, char *** rep, int ** pos, int ** cut)
+{
+  int i, j;
+  for (i = 0, j = 0; i < l; i++, j++) {
+    if (hyphens[i]&1) {
+      hyphword[j] = word[i];
+      if (*rep && *pos && *cut && (*rep)[i]) {
+        strcpy(hyphword + j - (*pos)[i] + 1, (*rep)[i]);
+        j += strlen((*rep)[i]) - (*pos)[i];
+        i += (*cut)[i] - (*pos)[i];
+      } else hyphword[++j] = '=';
+    } else hyphword[j] = word[i];
+  }
+  hyphword[j] = '\0';
+}
+
+
+/* main api function with default hyphenmin parameters */
+int hnj_hyphen_hyphenate2 (HyphenDict *dict,
+			   const char *word, int word_size, char * hyphens,
+			   char *hyphword, char *** rep, int ** pos, int ** cut)
+{
+  hnj_hyphen_hyph_(dict, word, word_size, hyphens, rep, pos, cut,
+    dict->clhmin, dict->crhmin, 1, 1);
+  hnj_hyphen_lhmin(dict->utf8, word, word_size,
+    hyphens, rep, pos, cut, (dict->lhmin > 0 ? dict->lhmin : 2));
+  hnj_hyphen_rhmin(dict->utf8, word, word_size,
+    hyphens, rep, pos, cut, (dict->rhmin > 0 ? dict->rhmin : 2));
+  if (hyphword) hnj_hyphen_hyphword(word, word_size, hyphens, hyphword, rep, pos, cut);
+  if (dict->utf8) return hnj_hyphen_norm(word, word_size, hyphens, rep, pos, cut);
+  return 0;
+}
+
+/* previous main api function with hyphenmin parameters */
+int hnj_hyphen_hyphenate3 (HyphenDict *dict,
+	const char *word, int word_size, char * hyphens,
+	char *hyphword, char *** rep, int ** pos, int ** cut,
+	int lhmin, int rhmin, int clhmin, int crhmin)
+{
+  lhmin = (lhmin > 0 ? lhmin : dict->lhmin);
+  rhmin = (rhmin > 0 ? rhmin : dict->rhmin);
+  hnj_hyphen_hyph_(dict, word, word_size, hyphens, rep, pos, cut,
+    clhmin, crhmin, 1, 1);
+  hnj_hyphen_lhmin(dict->utf8, word, word_size, hyphens,
+    rep, pos, cut, (lhmin > 0 ? lhmin : 2));
+  hnj_hyphen_rhmin(dict->utf8, word, word_size, hyphens,
+    rep, pos, cut, (rhmin > 0 ? rhmin : 2));
+  if (hyphword) hnj_hyphen_hyphword(word, word_size, hyphens, hyphword, rep, pos, cut);
+  if (dict->utf8) return hnj_hyphen_norm(word, word_size, hyphens, rep, pos, cut);
+  return 0;
+}
